/******************************************************************************
 * @(#)pl_proc.c : kernel board level infomation for procfs
 *
 * Copyright (c) 2002 Prolific Technology Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Prolific
 * Technology, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Prolific.
 *
 * History:
 *	2002-07-24	create file
 *			Jedy <jedy-wei@prolific.com.tw>
 *
 *	2005-03-07	to add PL-1061 clocking system
 *
 ******************************************************************************/
#include <linux/config.h>

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/pci.h>

#include <asm/hardware.h>


/*
 * definitions
 */

#define CLK_MAIN_OSCI_CONFIG_MASK       ((0x7) << 24)
#define CLK_MAIN_OSCI_EXT               0x4
#define CCR_OSC(x)                      (((x) & (CLK_MAIN_OSCI_CONFIG_MASK)) >> 24)
#define CCR_MPLLEN(x)                   ((x) & (1 << 11))
#define CCR_MPLLLD(x)                   ((x) & (0x1f))
#define CCR_FCLK2X(x)                   ((x) & (1 << 20))
#define CCR_CPUSRC(x)                   ((x) & (1 << 23))
#define CCR_PCICD(x)                    (((x) >> 16) & (0xf))
#define CCR_S_USB48(x)                  ((x) & (1 << 21))
#define CCR_USBCD(x)                    (((x) >> 28) & 0x7)
#define CCR_PCIDESKEW(x)                ((x) & (1 << 15))
#define CCR_EXTPCI(x)                   ((x) & (1 << 27))
#define CCR_PCIDRIVE(x)                 ((x) & (1 << 12))

#define CCR3_DEVCD(x)                   ((x) & 0xf)
#define CCR3_LCDCD(x)                   (((x) >> 4) & 0xf)
/*
 * Various clock divider types
 */
#define CD_CPU		0
#define CD_PCI		1
#define CD_DEV		2
#define CD_USB		3
#define CD_LCD		4
#define CD_MEM          5


/* default clock is 12MHz */
#define DEFAULT_CLK	    (12*1000)


/*
 * variables
 */
static unsigned int osci_freq = DEFAULT_CLK;    /* unit kHz */
static unsigned int xclk_in;    /* unit Hz */
static unsigned int pll_out;    /* unit Hz */



/*
 * forward referneces
 */
static unsigned int pl_get_clock(int which);
static void	    pl_set_clock(int which, unsigned int cd);


/*
 * get clock function unit kHz
 */
unsigned int
pl_get_cpu_clock(void)
{
    return pl_get_clock(CD_CPU)/1000;
} /* pl_get_cpu_clock() */

unsigned int
pl_get_pci_clock(void)
{
    return pl_get_clock(CD_PCI)/1000;
} /* pl_get_pci_clock() */

unsigned int
pl_get_dev_clock(void)
{
    return pl_get_clock(CD_DEV)/1000;
} /* pl_get_dev_clock() */

unsigned int
pl_get_lcd_clock(void)
{
    return pl_get_clock(CD_LCD)/1000;
} /* pl_get_lcd_clock() */

unsigned int
pl_get_usb_clock(void)
{
    return pl_get_clock(CD_USB)/1000;
} /* pl_get_usb_clock() */

unsigned int
pl_get_mem_clock(void)
{
    return pl_get_clock(CD_MEM)/1000;
} /* pl_get_usb_clock() */


/*
 * get clock function unit Hz
 */

unsigned int
pl_get_cpu_hz(void)
{
    return pl_get_clock(CD_CPU);
} /* pl_get_cpu_clock() */

unsigned int
pl_get_pci_hz(void)
{
    return pl_get_clock(CD_PCI);
} /* pl_get_pci_clock() */

unsigned int
pl_get_dev_hz(void)
{
    return pl_get_clock(CD_DEV);
} /* pl_get_dev_clock() */

unsigned int
pl_get_lcd_hz(void)
{
    return pl_get_clock(CD_LCD);
} /* pl_get_lcd_clock() */

unsigned int
pl_get_usb_hz(void)
{
    return pl_get_clock(CD_USB);
} /* pl_get_usb_clock() */

unsigned int
pl_get_mem_hz(void)
{
    return pl_get_clock(CD_MEM);
} /* pl_get_usb_clock() */


void
pl_set_lcd_clock(unsigned int cd)
{
    pl_set_clock(CD_LCD, cd);
} /* pl_set_lcd_clock() */



static unsigned int
pl_get_clock(int which)
{
    unsigned int ccr, ccr3, cd = 1;
    unsigned int clk;
    unsigned int xout;

    ccr  = readl(PL_CLK_CFG);
    ccr3 = readl(PL_CLK_CFG3);

    if (CCR_MPLLEN(ccr)) {
        xout = pll_out * 2;
    } else {
        xout = xclk_in;
    }

    switch (which) {
    case CD_CPU:
        if (CCR_MPLLEN(ccr)) { /* pll enable */
            if (CCR_CPUSRC(ccr)) { /* cpu src is 2x or 3x pll_out ? */
                if (CCR_FCLK2X(ccr))
                    clk = pll_out * 2;
                else
                    clk = pll_out * 3;
            } else {
                clk = pll_out;
            }

        } else {
            clk = xclk_in;
        }
        return clk;

    case CD_MEM:
        if (CCR_MPLLEN(ccr)) { /* pll enable */
            clk = pll_out;
        } else {
            clk = xclk_in;
        }
        return clk;

    case CD_PCI:
        cd = CCR_PCICD(ccr) + 1;
        clk = xout / cd;
	return clk;

    case CD_DEV:
	cd = CCR3_DEVCD(ccr3) + 1;
        clk = xout / cd;
        return clk;

    case CD_USB:
        if (CCR_S_USB48(ccr))
            clk =  48000*1000;
        else {
            cd = CCR_USBCD(ccr) + 1;
            clk = xout / cd;
        }

        return clk;

    case CD_LCD:
        cd = CCR3_LCDCD(ccr3) + 1;
        clk = xout /cd;
        return clk;
    }

    return 0;
} /* pl_get_clock() */

static void
pl_set_clock(int which, unsigned int cd)
{
    unsigned int ccr, ccr3;

    ccr  = readl(PL_CLK_CFG);
    ccr3 = readl(PL_CLK_CFG3);

    switch (which) {
    case CD_USB:
	if (cd > 8) cd = 8;
	    cd--;

        ccr &= ~(0x7 << 28);
        ccr |= (cd << 28);
        writel(ccr, PL_CLK_CFG);

	break;
    case CD_LCD:
	if (cd > 16) cd = 16;
	    cd--;

        ccr3 &= ~(0xf << 4);
        ccr3 |= (cd << 4);
	writel(ccr3, PL_CLK_CFG3);
	break;
    }
} /* pl_set_clock() */


static unsigned int
get_sub_dev_clk(unsigned int dclk, unsigned int cd)
{
    switch(cd) {
    case 0:        return 0;
    case 1:        return 48*1000*1000;
    case 2:        return 24*1000*1000;
    case 3:        return 12*1000*1000;
    case 4:        return dclk;
    case 5:        return dclk/2;
    case 6:        return dclk/4;
    case 7:        return dclk/8;
    }
    return 0;
}

static unsigned int
get_sub_dev_clk_uart(unsigned int dclk, unsigned int cd)
{
    switch(cd) {
    case 0:        return 0;
    case 1:        return dclk;
    case 2:        return dclk/2;
    case 3:        return dclk/4;
    case 4:        return 48*1000*1000;
    case 5:        return 24*1000*1000;
    case 6:        return 16*1000*1000;
    case 7:        return 8*1000*1000;
    }
    return 0;
}

static unsigned int
get_sub_dev_clk_ms(unsigned int dclk, unsigned int cd)
{
    switch(cd) {
    case 0:        return 0;
    case 1:        return 40*1000*1000;
    case 2:        return 20*1000*1000;
    case 3:        return 10*1000*1000;
    case 4:        return dclk;
    case 5:        return dclk/2;
    case 6:        return dclk/4;
    case 7:        return dclk/8;
    }
    return 0;

}

int
get_hardware_list(char *buf)
{
    unsigned int   rom;
    int	       y, m, d;
    unsigned int   t;
    unsigned short md[13] = { 0, 31, 59, 90, 120, 151, 181, 212,
				  243, 273, 304, 334, 365 };
    unsigned int   ccr, ccr3;
    unsigned int   cpu, pci;
    unsigned int   dev, lcd, usb, mem;
    int	       len = 0;
    unsigned int nand_cd, i2c_cd, ac97_cd, gpio_cd, ide1_cd;
    unsigned int sd_cd, pccard_cd, gpio2_cd, uart_cd, ms_cd, ide2_cd;

    /*
     * ROM code version information
     */
    // rom = readl(rROM_VER);
    rom = 0;

    y = (rom / 365) + 1970;
    t = (y / 4) - (1970 / 4) - ((y / 100) - (1970 / 100)) +
    (y / 400) - (1970 / 400);
    if ((y % 4) == 0) md[2] += 1;

    t = (rom - t) % 365;
    for (m = 1; m < 13; m++) if (t < md[m]) break;

    d = t - md[m-1] + 1;

    /*
     * reading other configuration values
     */
    ccr = readl(PL_CLK_CFG);
    ccr3 = readl(PL_CLK_CFG3);

    /*
     * information about the clocking system
     */
    cpu = pl_get_cpu_hz();
    pci = pl_get_pci_hz();

    dev = pl_get_dev_hz();
    usb = pl_get_usb_hz();
    lcd = pl_get_lcd_hz();
    mem = pl_get_mem_hz();

    /* Rom Version */
    len += sprintf(buf, "ROM Ver:   %d/%d/%d\n\n", y,m,d);

    /* Routing Detail */
    len += sprintf(buf+len, "Routing Detail\n"
        "Clock source: %s\n"
        "cfg_fdiv: %s\n"
        "sysclk_src: %s\n"
        "multiply_on: %s\n\n",
            (CCR_OSC(ccr) == CLK_MAIN_OSCI_EXT) ? "ext osc" : "usb phy (12MHz)",
            (CCR_FCLK2X(ccr)) ? "2x": "3x",
            (CCR_CPUSRC(ccr)) ? (CCR_CPUSRC(ccr) ? "2x" : "3x") : "1x",
            (CCR_MPLLEN(ccr)) ? "pllm" : "xclk_in");

    /* Divider and Registers */
    len += sprintf(buf+len, "Divider and Register\n"
        "cfg_pdiv: %d\n"
        "cfg_ldiv: %d\n"
        "cfg_ddiv: %d\n"
        "cfg_udiv: %d\n"
        "External PCI: %s\n"
        "PCI Deskew: %s\n"
        "PCI Drive: %dmA\n"
        "mxlt_s_usb2: %s\n"
        "ccr:  %08x\n"
        "ccr3: %08x\n\n",
            CCR_PCICD(ccr), CCR3_LCDCD(ccr3), CCR3_DEVCD(ccr3), CCR_USBCD(ccr),
            CCR_EXTPCI(ccr) ? "enabled" : "disabled",
            CCR_PCIDESKEW(ccr) ? "enabled" : "disabled",
            CCR_PCIDRIVE(ccr) ? 8 : 4,
            CCR_S_USB48(ccr) ? "usb2 clock" : "dev clock",
            ccr, ccr3);

    /* Clock Detail */
    len += sprintf(buf+len, "Clock Detail\n"
        "XCLK_IN:    %8d kHz\n"
        "PLL_OUT:    %8d kHz\n"
        "CPU clock:  %8d kHz\n"
        "MEM clock:  %8d kHz\n"
        "PCI clock:  %8d kHz\n"
        "DEV clock:  %8d kHz\n"
        "USB clock:  %8d kHz\n"
        "LCD clock:  %8d kHz\n\n",
        xclk_in/1000, pll_out/1000,
        cpu/1000, mem/1000,
        pci/1000, dev/1000, usb/1000,
        lcd/1000);


    nand_cd = readb(PL_CLK_NAND)& 0x7;
    i2c_cd = readb(PL_CLK_I2C)& 0x7;
    ac97_cd = readb(PL_CLK_AC97)& 0x7;
    gpio_cd = readb(PL_CLK_GPIO)& 0x7;
    ide1_cd = readb(PL_CLK_IDE1)& 0x7;
    sd_cd = readb(PL_CLK_SD)& 0x7;
    pccard_cd = readb(PL_CLK_PCCARD)& 0x7;
    gpio2_cd = readb(PL_CLK_GPIO2)& 0x7;
    uart_cd = readb(PL_CLK_UART)& 0x7;
    ms_cd = readb(PL_CLK_MS)& 0x7;
    ide2_cd = readb(PL_CLK_IDE2)& 0x7;

    /* Device Clock Detail */
    len += sprintf(buf+len, "Device Clock Detail\n"
        "NAND:       %8d kHz\n"
        "I2C:        %8d kHz\n"
        "AC97:       %8d kHz\n"
        "GPIO:       %8d kHz\n"
        "IDE1:       %8d kHz\n"
        "SD:         %8d kHz\n"
        "PCCARD:     %8d kHz\n"
        "GPIO2:      %8d kHz\n"
        "UART:       %8d kHz\n"
        "MS:         %8d kHz\n"
        "IDE2:       %8d kHz\n\n",
        get_sub_dev_clk(dev, nand_cd)/1000,
        get_sub_dev_clk(dev, i2c_cd)/1000,
        get_sub_dev_clk(dev, ac97_cd)/1000,
        get_sub_dev_clk(dev, gpio_cd)/1000,
        get_sub_dev_clk(dev, ide1_cd)/1000,
        get_sub_dev_clk(dev, sd_cd)/1000,
        get_sub_dev_clk(dev, pccard_cd)/1000,
        get_sub_dev_clk(dev, gpio2_cd)/1000,
        get_sub_dev_clk_uart(dev, uart_cd)/1000,
        get_sub_dev_clk_ms(dev, ms_cd)/1000,
        get_sub_dev_clk(dev, ide2_cd)/1000);


	return len;
} /* get_hardware_list() */

int get_malloc(char *buf)
{
    return 0;
}

static int __init
init_osci_freq(void)
{
    unsigned int ccr;
    unsigned int ns;


    ccr = readl(PL_CLK_CFG);
    if (CCR_OSC(ccr) == CLK_MAIN_OSCI_EXT) { /* external osci */
        xclk_in = osci_freq * 1000;
    } else { /* usb phy 12MHz */
        xclk_in = 12000 * 1000 ;
    }

    /* calculate clock out */
    if (CCR_MPLLEN(ccr)) { /* pll enable */
        ns = CCR_MPLLLD(ccr);
        if (ns <= 6) {
            printk("Error: PLL multiplication pll loop divider assigned wrong value %d\n", ns);
            ns = 1;
        }
        pll_out = xclk_in * ns;
    } else {
        pll_out = 0;
    }

    return 0;
}

module_init(init_osci_freq);

static int __init
osci_freq_setup(char *str)
{
    int	     f = 0;
    unsigned int ccr;
    int	     pll_enable;
    unsigned int ms, ns;

    /*
     * to get input clock frequency on MXTL_IO
     */
    if (str != (char *) 0) {
        for (;*str != '\0';str++) f = f * 10 + (*str - '0');
            osci_freq = (f == 0) ? osci_freq : f;
    }

    return 1;
} /* osci_freq_setup() */

__setup("osci_freq=", osci_freq_setup);
